import abc
import asyncio
import threading
import time
from queue import Queue
from typing import Dict, Tuple
from uuid import UUID, uuid4


class MessageConsumer(abc.ABC):
    @abc.abstractmethod
    async def handle_message(self, headers: dict, body: bytes) -> None:
        """
        Handle a message with the given headers and body.
        :param headers: The message headers.
        :param body: The message body.
        """
        raise NotImplementedError("Subclasses must implement this method.")


class MessagePump:
    def __init__(self):
        self._consumers: Dict[UUID, MessageConsumer] = dict()
        self._messages: Queue[Tuple[dict, bytes]] = Queue()
        self._task = None
        self._stopped = False

    def register_consumer(self, consumer: MessageConsumer) -> UUID:
        """
        Register a new consumer.
        :param consumer: The consumer to register.
        :return: A unique identifier for the consumer.
        """
        consumer_id = uuid4()
        self._consumers[consumer_id] = consumer
        return consumer_id

    def unregister_consumer(self, consumer_id: UUID):
        """
        Unregister a consumer.
        :param consumer_id: The unique identifier of the consumer to unregister.
        """
        if consumer_id in self._consumers:
            del self._consumers[consumer_id]

    def publish(self, message: Tuple[dict, bytes]):
        self._messages.put(message)

    def start(self):
        if self._task:
            return
        self._task = threading.Thread(target=self._consume, daemon=True)
        self._task.start()

    def _consume(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        while not self._stopped:
            if not self._messages.empty():
                headers, body = self._messages.get()
                loop.run_until_complete(
                    asyncio.gather(
                        *[
                            consumer.handle_message(headers, body)
                            for consumer in self._consumers.values()
                        ]
                    )
                )
                logging.info(f"Processed message with headers: {headers} and body: {body}")
            else:
                time.sleep(0.1)

    def stop(self):
        """
        Stop the message pump.
        """
        self._stopped = True
        if self._task:
            self._task.join(timeout=5)
            self._task = None



from typing import Tuple, List

from aett.eventstore import BaseEvent
from cloudevents.pydantic import CloudEvent

from sirabus import IHandleEvents
from sirabus.hierarchical_topicmap import HierarchicalTopicMap
from sirabus.servicebus import ServiceBus
from sirabus.servicebus.inmemory_servicebus import InMemoryServiceBus
from sirabus.message_pump import MessagePump


def _transform_cloudevent_message(
    topic_map: HierarchicalTopicMap, properties: dict, body: bytes
) -> Tuple[dict, BaseEvent]:
    ce = CloudEvent.model_validate_json(body)
    event_type = topic_map.resolve_type(ce.type)
    if event_type is None:
        raise ValueError(f"Event type {ce.type} not found in topic map")
    if event_type and not issubclass(event_type, BaseEvent):
        raise TypeError(f"Event type {event_type} is not a subclass of BaseModel")
    event = event_type.model_validate(ce.data)
    return properties, event


def create_servicebus_for_amqp_cloudevent(
    amqp_url: str,
    topic_map: HierarchicalTopicMap,
    handlers: List[IHandleEvents],
    prefetch_count: int = 10,
) -> ServiceBus:
    from sirabus.servicebus.amqp_servicebus import AmqpServiceBus

    return AmqpServiceBus(
        amqp_url=amqp_url,
        topic_map=topic_map,
        handlers=handlers,
        prefetch_count=prefetch_count,
        message_reader=_transform_cloudevent_message,
    )


def create_servicebus_for_memory_cloudevent(
    topic_map: HierarchicalTopicMap,
    handlers: List[IHandleEvents],
    message_pump: MessagePump,
) -> ServiceBus:
    return InMemoryServiceBus(
        topic_map=topic_map,
        handlers=handlers,
        message_reader=_transform_cloudevent_message,
        message_pump=message_pump,
    )



import asyncio
from typing import Callable, List, Tuple

from aett.eventstore import BaseEvent

from sirabus import IHandleEvents
from sirabus.message_pump import MessageConsumer, MessagePump
from sirabus.hierarchical_topicmap import HierarchicalTopicMap
from sirabus.servicebus import ServiceBus


class InMemoryServiceBus(ServiceBus, MessageConsumer):
    def __init__(
        self,
        topic_map: HierarchicalTopicMap,
        message_reader: Callable[
            [HierarchicalTopicMap, dict, bytes], Tuple[dict, BaseEvent]
        ],
        handlers: List[IHandleEvents],
        message_pump: MessagePump,
    ) -> None:
        super().__init__(
            topic_map=topic_map, message_reader=message_reader, handlers=handlers
        )
        self._message_pump = message_pump
        self._subscription = None

    async def handle_message(self, headers: dict, body: bytes) -> None:
        await self.handle_event(headers, body)

    async def run(self):
        if not self._subscription:
            self._subscription = self._message_pump.register_consumer(self)
        await asyncio.sleep(0)

    async def stop(self):
        if self._subscription:
            self._message_pump.unregister_consumer(self._subscription)
        await asyncio.sleep(0)



import abc
import asyncio
from typing import Tuple, Callable, List

from aett.eventstore import BaseEvent

from sirabus import IHandleEvents
from sirabus.hierarchical_topicmap import HierarchicalTopicMap


class ServiceBus(abc.ABC):
    def __init__(
        self,
        topic_map: HierarchicalTopicMap,
        message_reader: Callable[
            [HierarchicalTopicMap, dict, bytes], Tuple[dict, BaseEvent]
        ],
        handlers: List[IHandleEvents],
    ) -> None:
        self._topic_map = topic_map
        self._message_reader = message_reader
        self._handlers = handlers

    @abc.abstractmethod
    async def run(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def stop(self):
        raise NotImplementedError()

    async def handle_event(self, headers: dict, body: bytes) -> None:
        headers, event = self._message_reader(self._topic_map, headers, body)
        if not isinstance(event, BaseEvent):
            raise TypeError(f"Expected event of type BaseEvent, got {type(event)}")
        await asyncio.gather(
            *[
                h.handle(event=event, headers=headers)
                for h in self._handlers
                if hasattr(type(h), "event_type") and isinstance(event, type(h).event_type)
            ],
            return_exceptions=True,
        )



import abc
from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar, get_args, Optional

from aett.eventstore import BaseEvent, Topic
from aett.eventstore.base_command import BaseCommand
from pydantic import BaseModel, Field

TEvent = TypeVar("TEvent", bound=BaseEvent, contravariant=True)
TCommand = TypeVar("TCommand", bound=BaseCommand, contravariant=True)


@Topic("command_response")
class CommandResponse(BaseModel):
    """
    Represents a response to a command.
    This class can be extended to provide specific response types.
    """

    success: bool = Field(
        default=True, description="Indicates if the command was successful"
    )
    message: Optional[str] = Field(
        default="",
        description="A message providing additional information about the command response",
    )

    def __repr__(self) -> str:
        return f"CommandResponse(success={self.success}, message='{self.message}')"


class IRouteCommands(ABC, Generic[TCommand]):
    """
    Interface for routing commands. The command router expects to receive replies to commands
    """

    @abstractmethod
    async def route(self, command: TCommand) -> CommandResponse:
        """
        Route a command.

        :param command: The command to route.
        :return: A CommandResponse indicating the success or failure of the command routing.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")


class IPublishEvents(ABC, Generic[TEvent]):
    """
    Interface for publishing events.
    """

    @abstractmethod
    async def publish(self, event: TEvent) -> None:
        """
        Publish an event.

        :param event: The event to publish.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")


class IHandleEvents(ABC, Generic[TEvent]):
    """
    Interface for handling events.
    """

    event_type: Type[TEvent]

    def __init_subclass__(cls, **kwargs):
        cls.event_type = get_args(cls.__orig_bases__[0])[0]

    @abstractmethod
    async def handle(self, event: TEvent, headers: dict) -> None:
        """
        Handle an event.

        :param event: The event to handle.
        :param headers: Additional headers associated with the event.
        :return: None
        """
        raise NotImplementedError("This method should be overridden by subclasses.")


def generate_vhost_name(name: str, version: str) -> str:
    """
    Generates a virtual host name based on the application name and version.
    :param name: The name of the application.
    :param version: The version of the application.
    :return: A string representing the virtual host name.
    """
    import hashlib

    h = hashlib.sha256(usedforsecurity=False)
    h.update(f"{name}_{version}".encode())
    return h.hexdigest()



import logging
from typing import Set

import aio_pika
from aio_pika.abc import AbstractRobustConnection, AbstractChannel, ExchangeType

from sirabus.hierarchical_topicmap import HierarchicalTopicMap


class TopographyBuilder:
    def __init__(self, amqp_url: str, topic_map: HierarchicalTopicMap) -> None:
        self.__amqp_url = amqp_url
        self.__topic_map = topic_map

    async def build(self) -> None:
        connection: AbstractRobustConnection = await aio_pika.connect_robust(
            url=self.__amqp_url
        )
        await connection.connect()
        channel: AbstractChannel = await connection.channel()
        await self._build_topography(channel=channel)
        logging.debug("Topography built and consumers registered.")

    async def _build_topography(self, channel: AbstractChannel) -> None:
        exchanges: Set[str] = set()
        await self._declare_exchanges(channel, exchanges)
        relationships = self.__topic_map.build_parent_child_relationships()
        for key in relationships:
            for child in relationships[key]:
                # Declare the child exchange if it does not exist
                if child not in exchanges:
                    await self._declare_exchange(
                        topic=child, channel=channel, exchanges=exchanges
                    )
                # Bind the child exchange to the parent exchange
                destination = await channel.get_exchange(child)
                bind_response = await destination.bind(
                    exchange=key, routing_key=f"{child}.#"
                )
                logging.debug(f"Bound {child} to {key} with response {bind_response}")
        await channel.close()

    async def _declare_exchanges(self, channel, exchanges):
        all_topics = set(self.__topic_map.get_all_hierarchical_topics())
        for topic in all_topics:
            await self._declare_exchange(
                topic=topic, channel=channel, exchanges=exchanges
            )

    @staticmethod
    async def _declare_exchange(
        topic: str, channel: AbstractChannel, exchanges: Set[str]
    ) -> None:
        await channel.declare_exchange(
            name=topic, type=ExchangeType.TOPIC, durable=True
        )
        exchanges.add(topic)



import asyncio
import logging
import uuid

import aio_pika
from aett.eventstore import BaseEvent, Topic
from aio_pika import Message
from cloudevents.pydantic import CloudEvent
from pydantic import BaseModel, Field

from sirabus import IPublishEvents, TEvent
from sirabus.message_pump import MessagePump
from sirabus.hierarchical_topicmap import HierarchicalTopicMap


def create_cloud_event(
    event: TEvent, topic_map: HierarchicalTopicMap, logger
) -> (str, str, str):
    if not isinstance(event, BaseEvent):
        logger.exception(f"{event} is not an instance of BaseEvent", exc_info=True)
        raise TypeError(f"Expected event of type {type(TEvent)}, got {type(event)}")
    event_type = type(event)
    topic = Topic.get(event_type)
    hierarchical_topic = topic_map.get_hierarchical_topic(event_type)

    if not hierarchical_topic:
        raise ValueError(
            f"Topic for event type {event_type} not found in hierarchical_topic map."
        )
    a = CloudEventAttributes(
        id=str(uuid.uuid4()),
        specversion="1.0",
        datacontenttype="application/json",
        time=event.timestamp.isoformat(),
        source=event.source,
        subject=topic,
        type=hierarchical_topic or topic,
    )
    ce = CloudEvent(
        attributes=a.model_dump(),
        data=event.model_dump(mode="json"),
    )
    j = ce.model_dump_json()
    return topic, hierarchical_topic, j


class CloudEventAttributes(BaseModel):
    id: str = Field(default=str(uuid.uuid4()))
    specversion: str = Field(default="1.0")
    datacontenttype: str = Field(default="application/json")
    time: str = Field()
    source: str = Field()
    subject: str = Field()
    type: str = Field()


class AmqpCloudEventPublisher(IPublishEvents[TEvent]):
    """
    Publishes events as CloudEvents.
    """

    def __init__(
        self,
        amqp_url: str,
        topic_map: HierarchicalTopicMap,
        logger: logging.Logger | None = None,
    ) -> None:
        self.__amqp_url = amqp_url
        self.__topic_map = topic_map
        self.__logger = logger or logging.getLogger("CloudEventPublisher")

    async def publish(self, event: TEvent) -> None:
        """
        Publishes the event to the configured topic.
        :param event: The event to publish.
        """
        topic, hierarchical_topic, j = create_cloud_event(
            event, self.__topic_map, self.__logger
        )
        connection = await aio_pika.connect_robust(url=self.__amqp_url)
        channel = await connection.channel()
        exchange = await channel.get_exchange(name="amq.topic", ensure=True)
        self.__logger.debug("Channel opened for publishing CloudEvent.")
        response = await exchange.publish(
            message=Message(body=j.encode(), headers={"topic": topic}),
            routing_key=hierarchical_topic,
        )
        self.__logger.debug(f"Published {response}")
        await channel.close()
        await connection.close()


class InMemoryCloudEventPublisher(IPublishEvents[TEvent]):
    """
    Publishes events as CloudEvents in memory.
    """

    def __init__(
        self,
        topic_map: HierarchicalTopicMap,
        messagepump: MessagePump,
        logger: logging.Logger | None = None,
    ) -> None:
        self.__topic_map = topic_map
        self.__messagepump = messagepump
        self.__logger = logger or logging.getLogger("InMemoryCloudEventPublisher")

    async def publish(self, event: TEvent) -> None:
        """
        Publishes the event to the configured topic in memory.
        :param event: The event to publish.
        """
        topic, hierarchical_topic, j = create_cloud_event(
            event, self.__topic_map, self.__logger
        )
        self.__messagepump.publish(({"topic": topic}, j.encode()))
        await asyncio.sleep(0)


def create_publisher_for_amqp_cloudevent(
    amqp_url: str, topic_map: HierarchicalTopicMap, logger: logging.Logger | None = None
) -> IPublishEvents[BaseEvent]:
    """
    Creates a CloudEventPublisher for AMQP.
    :param amqp_url: The AMQP URL.
    :param topic_map: The hierarchical topic map.
    :param logger: Optional logger.
    :return: A CloudEventPublisher instance.
    """
    return AmqpCloudEventPublisher(
        amqp_url=amqp_url, topic_map=topic_map, logger=logger
    )


def create_publisher_for_memory_cloudevent(
    topic_map: HierarchicalTopicMap,
    messagepump: MessagePump,
    logger: logging.Logger | None = None,
) -> IPublishEvents[BaseEvent]:
    """
    Creates a CloudEventPublisher for in-memory use.
    :param topic_map: The hierarchical topic map.
    :param logger: Optional logger.
    :return: A CloudEventPublisher instance.
    """
    return InMemoryCloudEventPublisher(
        topic_map=topic_map, messagepump=messagepump, logger=logger
    )



import inspect
from typing import Dict, Set, Self, Any, Iterable

from aett.eventstore import Topic, BaseEvent
from pydantic import BaseModel


class HierarchicalTopicMap:
    """
    Represents a map of topics to event classes.
    """

    def __init__(self) -> None:
        self.__topics: Dict[str, type] = {}
        self.__excepted_bases__: Set[type] = {object, BaseModel, BaseEvent}

    def add(self, topic: str, cls: type) -> Self:
        """
        Adds the topic and class to the map.
        :param topic: The topic of the event.
        :param cls: The class of the event.
        """
        self.__topics[topic] = cls
        return self

    def except_base(self, t: type) -> None:
        """
        Exclude the base class from the topic hierarchy.
        :param t: The class to exclude.
        """
        if not isinstance(t, type):
            raise TypeError("Expected a class type")
        if t not in self.__excepted_bases__:
            self.__excepted_bases__.add(t)

    def register(self, instance: Any) -> Self:
        t = instance if isinstance(instance, type) else type(instance)
        topic = Topic.get(t)
        if topic not in self.__topics:
            self.add(topic, t)

        return self

    def _resolve_topics(self, t: type, suffix: str | None = None) -> str:
        topic = t.__topic__ if hasattr(t, "__topic__") else t.__name__
        # yield topic if suffix is None else f"{topic}.{suffix}"
        if any(tb for tb in t.__bases__ if tb not in self.__excepted_bases__):
            tbase = self._resolve_topics(t.__bases__[0], suffix)
            topic = (
                f"{tbase}.{topic}" if suffix is None else f"{tbase}.{topic}.{suffix}"
            )
            return topic
        return topic

    def register_module(self, module: object) -> Self:
        """
        Registers all the classes in the module.
        """
        for _, o in inspect.getmembers(module, inspect.isclass):
            if inspect.isclass(o):
                self.register(o)
            if inspect.ismodule(o):
                self.register_module(o)
        return self

    def resolve_type(self, topic: str) -> type | None:
        """
        Gets the class of the event given the topic.
        :param topic: The topic of the event.
        :return: The class of the event.
        """
        return self.__topics.get(topic, None)

    def get_hierarchical_topic(self, instance: type) -> str | None:
        """
        Gets the topic of the event given the class.
        :param instance: The class of the event.
        :return: The topic of the event.
        """
        t = instance if isinstance(instance, type) else type(instance)
        if t in self.__topics.values():
            n = self._resolve_topics(t)
            return n
        return None

    def get_all_hierarchical_topics(self) -> Iterable[str]:
        """
        Gets all the hierarchical topics in the map.
        :return: A list of all the hierarchical topics.
        """
        for topic in self.__topics.values():
            yield self._resolve_topics(topic)

    def build_parent_child_relationships(self) -> Dict[str, Set[str]]:
        """
        Builds a list of parent-child relationships for the given topic.
        :return: A list of parent-child relationships.
        """

        relationships: Dict[str, Set[str]] = {}

        def visit(cls: type) -> None:
            for base in cls.__bases__:
                if base not in self.__excepted_bases__:
                    parent_type = self.resolve_type(Topic.get(base))
                    if not parent_type:
                        raise RuntimeError(
                            f"Base class '{base.__name__}' for '{cls.__name__}' not found in the topic map."
                        )
                    parent = self.get_hierarchical_topic(parent_type)
                    if not parent:
                        raise RuntimeError(
                            f"Parent topic for class '{cls.__name__}' not found in the topic map."
                        )
                    child_type = self.resolve_type(Topic.get(cls))
                    if not child_type:
                        raise RuntimeError(
                            f"Child class '{cls.__name__}' not found in the topic map."
                        )
                    child = self.get_hierarchical_topic(child_type)
                    if not child:
                        raise RuntimeError(
                            f"Child topic for class '{cls.__name__}' not found in the topic map."
                        )
                    relationships.setdefault(parent, set()).add(child)
                    visit(base)

        for topic in self.__topics.values():
            if any(t for t in topic.__bases__ if t in self.__excepted_bases__):
                relationships.setdefault("amq.topic", set()).add(Topic.get(topic))
            visit(topic)
        return relationships



import logging
from typing import List, Optional, Set, Callable, Tuple

import aio_pika
from aett.eventstore import BaseEvent
from aio_pika.abc import (
    AbstractIncomingMessage,
    AbstractRobustConnection,
    AbstractRobustChannel,
)

from sirabus import IHandleEvents
from sirabus.hierarchical_topicmap import HierarchicalTopicMap
from sirabus.servicebus import ServiceBus


class AmqpServiceBus(ServiceBus):
    """
    Base class for event handlers.
    """

    def __init__(
        self,
        amqp_url: str,
        topic_map: HierarchicalTopicMap,
        handlers: List[IHandleEvents],
        message_reader: Callable[
            [HierarchicalTopicMap, dict, bytes], Tuple[dict, BaseEvent]
        ],
        prefetch_count: int = 10,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """Create a new instance of the consumer class, passing in the AMQP
        URL used to connect to RabbitMQ.
        :param str amqp_url: The AMQP URL to connect to RabbitMQ.
        :param HierarchicalTopicMap topic_map: The topic map to use for resolving topics.
        :param List[IHandleEvents] handlers: The list of event handlers to register.
        :param int prefetch_count: The number of messages to prefetch from RabbitMQ.
        """
        super().__init__(
            message_reader=message_reader, topic_map=topic_map, handlers=handlers
        )
        self.__handlers: List[IHandleEvents] = [handler for handler in handlers]
        self.__amqp_url = amqp_url
        self._topic_map = topic_map
        self._prefetch_count = prefetch_count
        self._logger = logger or logging.getLogger("ServiceBus")
        self.__topics = set(
            topic
            for topic in (
                self._topic_map.get_hierarchical_topic(type(handler).event_type)
                for handler in handlers
                if isinstance(handler, IHandleEvents)
            )
            if topic is not None
        )
        self.__queue_name = self._get_consumer_queue_name(self.__topics)
        self.__connection: Optional[AbstractRobustConnection] = None
        self.__channel: Optional[AbstractRobustChannel] = None
        self.__consumer_tag: Optional[str] = None

    async def __inner_handle_message(self, msg: AbstractIncomingMessage):
        try:
            await self.handle_event(msg.headers, msg.body)
            await msg.ack()
        except Exception as e:
            logging.exception("Exception while handling message", exc_info=e)
            await msg.nack(requeue=True)

    async def run(self):
        logging.debug("Starting service bus")
        self.__connection = await aio_pika.connect_robust(url=self.__amqp_url)
        self.__channel = await self.__connection.channel()
        await self.__channel.set_qos(prefetch_count=self._prefetch_count)
        logging.debug("Channel opened for consuming messages.")
        queue = await self.__channel.declare_queue(self.__queue_name, exclusive=True)
        for topic in self.__topics:
            await queue.bind(exchange=topic, routing_key=topic)
            logging.debug(f"Queue {self.__queue_name} bound to topic {topic}.")
        self.__consumer_tag = await queue.consume(callback=self.__inner_handle_message)

    async def stop(self):
        if self.__consumer_tag:
            queue = await self.__channel.get_queue(self.__queue_name)
            await queue.cancel(self.__consumer_tag)
            await self.__channel.close()
            await self.__connection.close()

    @staticmethod
    def _get_consumer_queue_name(topics: Set[str]) -> str:
        """
        Returns the queue name for the given topic.
        :param topics: The topics for which to get the queue name.
        :return: The queue name.
        """
        import hashlib

        h = hashlib.sha256(usedforsecurity=False)
        for topic in topics:
            h.update(topic.encode())
        hashed_topics = h.hexdigest()
        return hashed_topics
